@@ -1,13 +1,14 @@
 #!/usr/bin/env python3
+import pickle
 import requests
 import json
-import warnings
 import re
 import time
 from datetime import timedelta
 import math
 import argparse
 from ..default_servers import get_default_servers
+import os
 
 
 def cli_credentials():
@@ -74,7 +75,7 @@ def chunks(l, n):
 
 
 class Neo4jConnect:
-    """Thin layer over REST API to hold connection details, 
+    """Thin layer over REST API to hold connection details,
     handle multi-statement POST queries, return results and report errors.
 
     :param endpoint: a neo4j REST endpoint
@@ -95,8 +96,9 @@ class Neo4jConnect:
             print("Falling back to Neo4j v3 connection")
             self.commit = "/db/data/transaction/commit"
             self.headers = {}
-            self.test_connection()
-       
+            if not self.test_connection():
+                raise Exception("Failed to connect to Neo4j.")
+
     def commit_list(self, statements, return_graphs=False):
         """Commit a list of statements to neo4J DB via REST API.
         Errors prompt warnings (STDERR), not exceptions, and cause return = FALSE.
@@ -110,19 +112,18 @@ class Neo4jConnect:
         if return_graphs:
             for s in statements:
                 cstatements.append({'statement': s, "resultDataContents" : [ "row", "graph" ]})
-        else:        
+        else:
             for s in statements:
                 cstatements.append({'statement': s}) # rows an columns are returned by default.
         payload = {'statements': cstatements}
-        response = requests.post(url = "%s%s" 
+        response = requests.post(url = "%s%s"
                                  % (self.base_uri, self.commit), auth = (self.usr, self.pwd) ,
                                   data = json.dumps(payload), headers = self.headers)
         if self.rest_return_check(response):
             return response.json()['results']
         else:
             return False
-        
-        
+
     def commit_list_in_chunks(self, statements, verbose=False, chunk_length=1000):
 
         """Commit multiple (chunked) commit of statements to neo4J DB via REST API.
@@ -142,7 +143,7 @@ class Neo4jConnect:
             if verbose:
                 start_time = time.time()
                 print("Processing chunk of %d of %d starting with: %s" % (i,
-                                                                          c_no, 
+                                                                          c_no,
                                                                           c[0].encode('utf8')))
             r = self.commit_list(c)
             if verbose:
@@ -164,22 +165,22 @@ class Neo4jConnect:
         self.commit_list([cypher])
 
     def rest_return_check(self, response):
-        """Checks status response to post. Prints warnings to STDERR if not OK.
-        If OK, checks for errors in response. Prints any present as warnings to STDERR.
+        """Checks status response to post. Prints warnings if not OK.
+        If OK, checks for errors in response. Prints any present as warnings.
         Returns True STATUS OK and no errors, otherwise returns False.
         """
         if not (response.status_code == 200):
-            warnings.warn("Connection error: %s (%s)" % (response.status_code, response.reason))
+            print("\033[31mConnection Error:\033[0m %s (%s)" % (response.status_code, response.reason))
             return False
         else:
             j = response.json()
             if j['errors']:
                 for e in j['errors']:
-                    warnings.warn(str(e))
+                    print("\033[31mQuery Error:\033[0m " + str(e))
                 return False
             else:
                 return True
-            
+
     def test_connection(self):
         """Test neo4j endpoint connection"""
         statements = ["MATCH (n) RETURN n LIMIT 1"]
@@ -187,61 +188,165 @@ class Neo4jConnect:
             return True
         else:
             return False
-        
+
     def list_all_node_props(self):
         r = self.commit_list(['MATCH (n) with keys(n) AS kl UNWIND kl as k RETURN DISTINCT k'])
         d = dict_cursor(r)
         return [x['k'] for x in d]
-    
+
     def list_all_edge_props(self):
         r = self.commit_list(['MATCH ()-[r]-() with keys(r) AS kl UNWIND kl as k RETURN DISTINCT k'])
         d = dict_cursor(r)
         return [x['k'] for x in d]
 
-    def get_lookup(self, limit_by_prefix=None, include_individuals=False,
-                   limit_properties_by_prefix=('RO', 'BFO', 'VFBext')):
-
+    def get_lookup(self, cache=None, limit_type_by_prefix=('FBbt', 'VFBexp', 'VFBext'), include_individuals=True,
+               limit_properties_by_prefix=('RO', 'BFO', 'VFBext'), curies=False, include_synonyms=True, verbose=False):
         """Generate a name:ID lookup from a VFB neo4j DB, optionally restricted by a list of prefixes.
 
+        :param use_cache: If `True`, uses a cached lookup. If `False`, generates a new lookup.
         :param limit_by_prefix:  Optional list of id prefixes for limiting lookup of classes & individuals
         :param include_individuals: If `True`, individuals included in lookup.
         :param limit_properties_by_prefix:  Optional list of id prefixes for limiting lookup of properties.
-
+        :param curies: If `True`, returns CURIEs instead of IDs.
+        :param include_synonyms: If `True`, includes synonyms in the lookup.
+        :return: A dictionary with names (or synonyms) as keys and their corresponding IDs as values.
         """
-        if limit_by_prefix:
-            regex_string = '.+|'.join(limit_by_prefix) + '.+'
-            where = " WHERE a.short_form =~ '%s' " % regex_string
-        else:
-            where = ''
-        neo_labels = ['Class']
-        if include_individuals:
-            neo_labels.append('Individual')
+        try:
+            three_months_in_seconds = 3 * 30 * 24 * 60 * 60
+            if cache and os.path.exists(cache) and os.path.getctime(cache) > time.time() - three_months_in_seconds:
+                print("Loading cache lookup from disk...") if verbose else None
+                with open(cache, 'rb') as f:
+                    return pickle.load(f)
+        except Exception as e:
+            print(f"Failed to load cache lookup from disk: {e}")
+
+        # Print the loading message
+        print("Caching all terms for faster lookup...")
+
+        where = ''
+        if limit_type_by_prefix:
+            regex_string = '.+|'.join(limit_type_by_prefix) + '.+'
+            where += "AND a.short_form =~ '%s' " % regex_string
+        where += "AND NOT a:Deprecated "
+
         out = []
-        for l in neo_labels:
-            lookup_query = "MATCH (a:%s) %s RETURN a.short_form as id, a.label as name" % (l, where)
+        # All Classes wanted, where id starts with prefix in limit_type_by_prefix
+        l = 'Class'
+        print(f"Caching {l} names...")
+        lookup_query = "MATCH (a:%s) WHERE EXISTS(a.short_form) AND EXISTS(a.label) %s RETURN a.short_form as id, a.label as name" % (l, where)
+        q = self.commit_list([lookup_query])
+        out.extend(dict_cursor(q))
+        print(f"Caching {l} symbols...")
+        lookup_query = "MATCH (a:%s) WHERE EXISTS(a.short_form) %s AND exists(a.symbol) " \
+                    "RETURN a.short_form as id, a.symbol[0] as name" % (l, where)
+        q = self.commit_list([lookup_query])
+        out.extend(dict_cursor(q))
+
+        if include_synonyms:
+            print(f"Caching {l} synonyms...")
+            lookup_query = "MATCH (a:%s) WHERE EXISTS(a.short_form) %s AND EXISTS(a.synonyms) OR (a)-[:has_reference {typ:'syn'}]->(:pub:Individual) " \
+                        "UNWIND a.synonyms AS synonym2 " \
+                        "RETURN DISTINCT a.short_form AS id, synonym2 AS name " \
+                        "UNION ALL MATCH (a)-[r:has_reference {typ:'syn'}]->(:pub:Individual) " \
+                        "UNWIND r.value AS synonym1 " \
+                        "WITH a.short_form AS id, synonym1 AS synonym " \
+                        "RETURN DISTINCT id, synonym AS name" % (l, where)
             q = self.commit_list([lookup_query])
             out.extend(dict_cursor(q))
-            lookup_query = "MATCH (a:%s) %s AND exists(a.symbol) " \
-                           "RETURN a.short_form as id, a.symbol[0] as name" % (l, where)
+
+        if include_individuals:
+            where = "AND NOT a:Deprecated AND NOT a.short_form STARTS WITH 'VFBc_' AND NOT a:Person "
+            # If individuals are wanted, get them
+            l = 'Individual'
+            print(f"Caching {l} names...")
+            lookup_query = "MATCH (a:%s) WHERE EXISTS(a.short_form) AND EXISTS(a.label) %s RETURN a.short_form as id, a.label as name" % (l, where)
             q = self.commit_list([lookup_query])
             out.extend(dict_cursor(q))
+            print(f"Caching {l} symbols...")
+            lookup_query = "MATCH (a:%s) WHERE EXISTS(a.short_form) %s AND exists(a.symbol) " \
+                        "RETURN a.short_form as id, a.symbol[0] as name" % (l, where)
+            q = self.commit_list([lookup_query])
+            out.extend(dict_cursor(q))
+
+            if include_synonyms:
+                print(f"Caching {l} synonyms...")
+                lookup_query = "MATCH (a:%s) WHERE EXISTS(a.short_form) %s AND EXISTS(a.synonyms) OR (a)-[:has_reference {typ:'syn'}]->(:pub:Individual) " \
+                            "UNWIND a.synonyms AS synonym2 " \
+                            "RETURN DISTINCT a.short_form AS id, synonym2 AS name " \
+                            "UNION ALL MATCH (a)-[r:has_reference {typ:'syn'}]->(:pub:Individual) " \
+                            "UNWIND r.value AS synonym1 " \
+                            "WITH a.short_form AS id, synonym1 AS synonym " \
+                            "RETURN DISTINCT id, synonym AS name" % (l, where)
+                q = self.commit_list([lookup_query])
+                out.extend(dict_cursor(q))
+
         # All ObjectProperties wanted, irrespective of ID
         if limit_properties_by_prefix:
-            regex_string = '.+|'.join(limit_properties_by_prefix) + '.+'
-            where = " WHERE a.short_form =~ '%s' " % regex_string
+            match_string = "' OR a.short_form STARTS WITH '".join(limit_properties_by_prefix)
+            where = " WHERE a.short_form STARTS WITH '%s' " % match_string
         else:
             where = ''
+        print(f"Caching ObjectProperties...")
         property_lookup_query = "MATCH (a:ObjectProperty) " \
-                                " " + where + " " \
+                                + where + \
                                 "RETURN a.short_form as id, a.label as name"
         q = self.commit_list([property_lookup_query])
         out.extend(dict_cursor(q))
-        lookup = {x['name']: x['id'].replace('_', ':') for x in out}
-        lookup.update({x['id']: x['id'].replace('_', ':') for x in out})
-        # print(lookup['neuron'])
+
+        # All alternative terms for ObjectProperties
+        if limit_properties_by_prefix:
+            match_string = "' OR a.short_form STARTS WITH '".join(limit_properties_by_prefix)
+            where = " WHERE exists(a.alternative_term) and size(a.alternative_term) > 0 and a.short_form STARTS WITH '%s' " % match_string
+        else:
+            where = ''
+        print(f"Caching ObjectProperties alternative labels...")
+        property_lookup_query = "MATCH (a:ObjectProperty) " \
+                                + where + \
+                                "UNWIND a.alternative_term as label " + \
+                                "RETURN a.short_form as id, label as name"
+        q = self.commit_list([property_lookup_query])
+
+        # Iterate through the results and add only if the label doesn't already exist to avoid duplicate object properties e.g. overlaps
+        name_to_id = {item['name']: item['id'] for item in out}
+        print(f"Initial lookup contains {len(name_to_id)} entries") if verbose else None
+        print(f"Result for overlaps: {name_to_id['overlaps']}") if verbose else None
+        existing_names = set(name_to_id.keys())
+        for result in dict_cursor(q):
+            if not result['name'] in existing_names:
+                out.append(result)
+                existing_names.add(result['name'])
+                name_to_id[result['name']] = result['id']
+            else:
+                print(f"Skipping duplicate object property: {result['name']} - {result['id']} in favour or existing {name_to_id[result['name']]}") if verbose else None
+
+        # Removing duplicates while maintaining order
+        seen = set()
+        unique_out = []
+
+        for item in out:
+            pair = (item['name'], item['id'])
+            if pair not in seen:
+                seen.add(pair)
+                unique_out.append(item)
+
+        # Construct the final lookup dictionary
+        if curies:
+            lookup = {x['name']: x['id'].replace('_', ':') for x in unique_out}
+            lookup.update({x['id'].replace(':', '_'): x['id'] for x in set(lookup.values())})
+        else:
+            lookup = {x['name']: x['id'] for x in unique_out}
+
+        # Save the lookup to a cache file
+        try:
+            if cache:
+                with open(cache, 'wb') as f:
+                    pickle.dump(lookup, f)
+        except Exception as e:
+            print(f"Failed to save cache lookup on disk: {e}")
         return lookup
 
 
+
 def dict_cursor(results):
     """Takes JSON results from a neo4J query and turns them into a list of dicts.
 
@@ -262,7 +367,7 @@ def escape_string(strng):
     if type(strng) == str:
         strng = re.sub(r'\\', r'\\\\', strng)
         strng = re.sub("'", "\\'", strng)
-        strng = re.sub('"', '\\"', strng)        
+        strng = re.sub('"', '\\"', strng)
     return strng
 
 
